// This is a generated file, do not edit it directly.

/* This file is part of brille {{revision}}.

{{copyright}}

brille is free software: you can redistribute it and/or modify it under the
terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option)
any later version.

brille is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with brille. If not, see <https://www.gnu.org/licenses/>.            */

// {{includes}}

#pragma ACME comments off
#pragma ACME disable DOXYGEN_GENERATING_OUTPUT

#pragma ACME path src
#pragma ACME path lib/tetgen

#include "version.hpp"
#include "bz_mesh.hpp"
#include "bz_nest.hpp"
#include "bz_trellis.hpp"
#include "hall_symbol.hpp"
